/*
 * This file is part of the DCStella project
 * location: dcstella/aica/TIASound.c
 * heavily modified by SvOlli
 */

#include "TIASound.h"
#include <stdio.h>
#include <stdlib.h>
#include <time.h>

#define FALSE       0
#define TRUE        1

#define USE_RANDOM  FALSE  /* TRUE will also work if the source of 
			      rand() will be supplied */

/* define some data types to keep it platform independent */

#define static /* not sure if static really works */

/* CONSTANT DEFINITIONS */

/* definitions for AUDCx (15, 16) */
#define SET_TO_1     0x00      /* 0000 */
#define POLY4        0x01      /* 0001 */
#define DIV31_POLY4  0x02      /* 0010 */
#define POLY5_POLY4  0x03      /* 0011 */
#define PURE         0x04      /* 0100 */
#define PURE2        0x05      /* 0101 */
#define DIV31_PURE   0x06      /* 0110 */
#define POLY5_2      0x07      /* 0111 */
#define POLY9        0x08      /* 1000 */
#define POLY5        0x09      /* 1001 */
#define DIV31_POLY5  0x0a      /* 1010 */
#define POLY5_POLY5  0x0b      /* 1011 */
#define DIV3_PURE    0x0c      /* 1100 */
#define DIV3_PURE2   0x0d      /* 1101 */
#define DIV93_PURE   0x0e      /* 1110 */
#define DIV3_POLY5   0x0f      /* 1111 */
                 
#define DIV3_MASK    0x0c                 
                 
#define AUDC0        0x15
#define AUDC1        0x16
#define AUDF0        0x17
#define AUDF1        0x18
#define AUDV0        0x19
#define AUDV1        0x1a

/* the size (in entries) of the 4 polynomial tables */
#define POLY4_SIZE  0x000f
#define POLY5_SIZE  0x001f
#define POLY9_SIZE  0x01ff

/* channel definitions */
#define CHAN1       0
#define CHAN2       1


/* LOCAL GLOBAL VARIABLE DEFINITIONS */

/* structures to hold the 6 tia sound control bytes */

unsigned long /*char*/ audc_0, audf_0, audv_0, audc_1, audf_1, audv_1;
unsigned long /*char*/ div_n_cnt_0, div_n_cnt_1, div_n_max_0, div_n_max_1;
unsigned long /*char*/ p4_0, p4_1, p5_0, p5_1, p9_0, p9_1, outvol_0, outvol_1;


/* Initialze the bit patterns for the polynomials. */

/* The 4bit and 5bit patterns are the identical ones used in the tia chip. */
/* Though the patterns could be packed with 8 bits per byte, using only a */
/* single bit per byte keeps the math simple, which is important for */
/* efficient processing. */

static unsigned long /*char*/ Bit4[POLY4_SIZE] =
      { 1,1,0,1,1,1,0,0,0,0,1,0,1,0,0 };

static unsigned long /*char*/ Bit5[POLY5_SIZE] =
      { 0,0,1,0,1,1,0,0,1,1,1,1,1,0,0,0,1,1,0,1,1,1,0,1,0,1,0,0,0,0,1 };

/* I've treated the 'Div by 31' counter as another polynomial because of */
/* the way it operates.  It does not have a 50% duty cycle, but instead */
/* has a 13:18 ratio (of course, 13+18 = 31).  This could also be */
/* implemented by using counters. */

static unsigned long /*char*/ Div31[POLY5_SIZE] =
      { 0,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,0,0,0,0,0,0,0,0,0,0,0 };

/* Rather than have a table with 511 entries, I use a random number */
/* generator. */

#if USE_RANDOM
static unsigned long /*char*/ Bit9[POLY9_SIZE];
#else
// precalculated table by running rand() & 0x01 for 511 times :-)
static unsigned long /*char*/ Bit9[POLY9_SIZE] =
{1,0,1,1,1,1,0,0,1,1,0,1,0,1,1,0,0,0,0,0,1,0,1,1,0,0,0,1,1,1,1,0,0,0,1,
 1,1,0,1,0,1,1,1,1,0,1,0,0,1,0,1,0,1,0,0,1,0,0,0,1,1,1,0,1,0,1,0,1,1,1,
 0,1,0,1,0,1,0,0,1,0,1,0,0,0,0,0,1,1,0,1,0,0,0,0,1,0,0,0,0,1,1,0,0,0,1,
 1,1,0,1,0,0,0,1,0,0,1,1,1,0,1,0,1,1,1,1,1,1,1,1,1,1,0,1,1,0,1,0,1,1,0,
 0,1,0,1,1,1,0,0,0,1,0,1,0,1,0,1,0,1,1,1,1,0,0,0,1,0,1,0,0,0,0,0,1,0,1,
 1,1,1,1,0,0,1,1,0,1,1,0,1,0,1,1,1,1,1,0,0,1,1,0,1,0,0,1,1,1,0,0,0,0,0,
 0,0,1,1,1,0,0,1,0,1,0,1,0,1,0,0,1,1,0,0,1,0,0,0,1,1,1,0,0,1,0,0,1,1,0,
 0,0,0,1,0,1,1,1,0,0,1,0,0,0,0,0,1,0,0,0,0,0,1,0,0,0,0,0,1,1,0,0,1,1,1,
 1,0,1,0,0,1,1,1,1,0,1,1,1,0,0,1,0,0,0,0,0,0,0,0,0,1,1,0,1,0,1,0,0,0,1,
 1,0,0,0,1,0,1,1,0,1,1,1,1,1,0,1,1,0,1,1,0,1,0,0,0,0,0,0,1,0,1,0,0,0,0,
 0,0,1,1,0,1,0,0,0,1,0,0,0,0,1,1,1,0,0,1,0,0,1,1,1,0,0,1,0,0,1,0,1,1,1,
 1,0,1,1,0,0,1,0,0,0,0,0,1,0,0,0,1,0,0,0,0,0,0,1,0,1,1,1,0,0,0,1,0,0,1,
 0,0,0,1,1,0,1,1,1,1,1,0,0,1,0,0,1,0,1,1,0,0,0,1,0,0,0,1,0,0,0,0,0,1,1,
 1,1,0,0,1,0,1,1,0,1,1,1,0,1,0,1,1,0,1,1,0,1,1,1,1,1,0,0,1,1,1,1,0,0,1,
 1,0,1,0,0,0,1,1,0,0,1,0,0,1,1,1,1,1,0,1,0};
#endif

/* In my routines, I treat the sample output as another divide by N counter. */
/* For better accuracy, the Samp_n_cnt has a fixed binary decimal point */
/* which has 8 binary digits to the right of the decimal point. */

//static unsigned short Samp_n_max; /* Sample max, multiplied by 256 */
static unsigned long Samp_n_cnt; /* Sample cnt. */

#define Samp_n_max 256

/*****************************************************************************/
/* Module:  Tia_sound_init()                                                 */
/* Purpose: to handle the power-up initialization functions                  */
/*          these functions should only be executed on a cold-restart        */
/*                                                                           */
/* Author:  Ron Fries                                                        */
/* Date:    September 10, 1996                                               */
/*                                                                           */
/* Inputs:  sample_freq - the value for the '30 Khz' Tia audio clock         */
/*          playback_freq - the playback frequency in samples per second     */
/*                                                                           */
/* Outputs: Adjusts local globals - no return value                          */
/*                                                                           */
/*****************************************************************************/

void Tia_sound_init (unsigned long sample_freq, unsigned long playback_freq)
{
  unsigned char chan;
#if USE_RANDOM
  short n;
  
  /* fill the 9bit polynomial with random bits */
  for (n=0; n<POLY9_SIZE; n++)
  {
    Bit9[n] = rand() & 0x01;       /* fill poly9 with random bits */
  }
#endif
  
  /* calculate the sample 'divide by N' value based on the playback freq. */
//  Samp_n_max = 8; //(sample_freq<<8)/playback_freq;
  Samp_n_cnt = 0;  /* initialize all bits of the sample counter */
  
  /* initialize the local globals */
  outvol_0 = 0;
  outvol_1 = 0;

  div_n_cnt_0 = 0;
  div_n_cnt_1 = 0;

  div_n_max_0 = 0;
  div_n_max_1 = 0;

  audc_0 = 0;
  audc_1 = 0;
  
  audf_0 = 0;
  audf_1 = 0;

  audv_0 = 0;
  audv_1 = 0;
  
  p4_0 = 0;
  p4_1 = 0;

  p5_0 = 0;
  p5_1 = 0;

  p9_0 = 0;
  p9_1 = 0;
}


void update_chan_0(void)
{
  unsigned long new_val;
  /* an AUDC value of 0 is a special case */
  if (audc_0 == SET_TO_1)
  {
    /* indicate the clock is zero so no processing will occur */
    new_val = 0;
    
    /* and set the output to the selected volume */
    outvol_0 = audv_0;
  }
  else
  {
    /* otherwise calculate the 'divide by N' value */
    new_val = audf_0 + 1;
    
    /* if bits 2 & 3 are set, then multiply the 'div by n' count by 3 */
    if ((audc_0 & DIV3_MASK) == DIV3_MASK)
    {
      new_val *= 3;
    }
  }
  
  /* only reset those channels that have changed */
  if (new_val != div_n_max_0)
  {
    /* reset the divide by n counters */
    div_n_max_0 = new_val;
    
    /* if the channel is now volume only or was volume only */
    if ((div_n_cnt_0 == 0) || (new_val == 0))
    {
      /* reset the counter (otherwise let it complete the previous) */
      div_n_cnt_0 = new_val;
    }
  }
}


void update_chan_1(void)
{
  unsigned long new_val;
  /* an AUDC value of 0 is a special case */
  if (audc_1 == SET_TO_1)
  {
    /* indicate the clock is zero so no processing will occur */
    new_val = 0;
    
    /* and set the output to the selected volume */
    outvol_1 = audv_1;
  }
  else
  {
    /* otherwise calculate the 'divide by N' value */
    new_val = audf_1 + 1;
    
    /* if bits 2 & 3 are set, then multiply the 'div by n' count by 3 */
    if ((audc_1 & DIV3_MASK) == DIV3_MASK)
    {
      new_val *= 3;
    }
  }
  
  /* only reset those channels that have changed */
  if (new_val != div_n_max_1)
  {
    /* reset the divide by n counters */
    div_n_max_1 = new_val;
    
    /* if the channel is now volume only or was volume only */
    if ((div_n_cnt_1 == 0) || (new_val == 0))
    {
      /* reset the counter (otherwise let it complete the previous) */
      div_n_cnt_1 = new_val;
    }
  }
}



/*****************************************************************************/
/* Module:  Tia_process()                                                    */
/* Purpose: To fill the output buffer with the sound output based on the     */
/*          tia chip parameters.  This routine has been optimized.           */
/*                                                                           */
/* Author:  Ron Fries                                                        */
/* Date:    September 10, 1996                                               */
/*                                                                           */
/* Inputs:  *buffer - pointer to the buffer where the audio output will      */
/*                    be placed                                              */
/*          n - size of the playback buffer                                  */
/*                                                                           */
/* Outputs: the buffer will be filled with n bytes of audio - no return val  */
/*                                                                           */
/*****************************************************************************/

void Tia_process (unsigned long address, unsigned long n)
{
  unsigned long current;
  unsigned long save32;
  
  /* loop until the buffer is filled */
  for(current=address; current<address+n; current+=2)
  {
    /* Process channel 0 */
    if (div_n_cnt_0 > 1)
    {
      div_n_cnt_0--;
    }
    else if (div_n_cnt_0 == 1)
    {
      div_n_cnt_0 = div_n_max_0;
      
      /* the P5 counter has multiple uses, so we inc it here */
      p5_0++;
      if (p5_0 >= POLY5_SIZE)
	p5_0 = 0;
      
      /* check clock modifier for clock tick */
      if  (((audc_0 & 0x02) == 0) ||
	   (((audc_0 & 0x01) == 0) && Div31[p5_0]) ||
	   (((audc_0 & 0x01) == 1) &&  Bit5[p5_0]))
      {
	if (audc_0 & 0x04)       /* pure modified clock selected */
	{
	  if (outvol_0)        /* if the output was set */
	    outvol_0 = 0;     /* turn it off */
	  else
	    outvol_0 = audv_0; /* else turn it on */
	}
	else if (audc_0 & 0x08)    /* check for p5/p9 */
	{
	  if (audc_0 == POLY9)    /* check for poly9 */
	  {
	    /* inc the poly9 counter */
	    p9_0++;
	    if (p9_0 >= POLY9_SIZE)
	      p9_0 = 0;
	    
	    if (Bit9[p9_0])
	      outvol_0 = audv_0;
	    else
	      outvol_0 = 0;
	  }
	  else                        /* must be poly5 */
	  {
	    if (Bit5[p5_0])
	      outvol_0 = audv_0;
	    else
	      outvol_0 = 0;
	  }
	}
	else  /* poly4 is the only remaining option */
	{
	  /* inc the poly4 counter */
	  p4_0++;
	  if (p4_0 >= POLY4_SIZE)
	    p4_0 = 0;
	  
	  if (Bit4[p4_0])
	    outvol_0 = audv_0;
	  else
	    outvol_0 = 0;
	}
      }
    }
    
    
    /* Process channel 1 */
    if (div_n_cnt_1 > 1)
    {
      div_n_cnt_1--;
    }
    else if (div_n_cnt_1 == 1)
    {
      div_n_cnt_1 = div_n_max_1;
      
      /* the P5 counter has multiple uses, so we inc it here */
      p5_1++;
      if (p5_1 >= POLY5_SIZE)
	p5_1 = 0;
      
      /* check clock modifier for clock tick */
      if  (((audc_1 & 0x02) == 0) ||
	   (((audc_1 & 0x01) == 0) && Div31[p5_1]) ||
	   (((audc_1 & 0x01) == 1) &&  Bit5[p5_1]))
      {
	if (audc_1 & 0x04)       /* pure modified clock selected */
	{
	  if (outvol_1)        /* if the output was set */
	    outvol_1 = 0;     /* turn it off */
	  else
	    outvol_1 = audv_1; /* else turn it on */
	}
	else if (audc_1 & 0x08)    /* check for p5/p9 */
	{
	  if (audc_1 == POLY9)    /* check for poly9 */
	  {
	    /* inc the poly9 counter */
	    p9_1++;
	    if (p9_1 >= POLY9_SIZE)
	      p9_1 = 0;
	    
	    if (Bit9[p9_1])
	      outvol_1 = audv_1;
	    else
	      outvol_1 = 0;
	  }
	  else                        /* must be poly5 */
	  {
	    if (Bit5[p5_1])
	      outvol_1 = audv_1;
	    else
	      outvol_1 = 0;
	  }
	}
	else  /* poly4 is the only remaining option */
	{
	  /* inc the poly4 counter */
	  p4_1++;
	  if (p4_1 >= POLY4_SIZE)
	    p4_1 = 0;
	  
	  if (Bit4[p4_1])
	    outvol_1 = audv_1;
	  else
	    outvol_1 = 0;
	}
      }
    }

    save32 |= (outvol_0 + outvol_1 + 0) << (((current & 3)+1) * 8);
    if ((current & 3) == 2)
    {
      *(unsigned long*)(current & ~3) = save32;
      save32 = 0;
    }
  }
}
