/*
 * This file is part of the DCStella project
 * location: dcstella/core/discio.c
 * written by SvOlli
 */

#include <unistd.h>

#include <dream.h>

#include "screen.h"
#include "discio.h"
#include "o/compilemode.h"

char *configkey[MAX_KEYS],*configvalue[MAX_KEYS];
char *filename[MAX_ROMS];
char *romname[MAX_ROMS];
int  numkeys;
char nullstr[]="";

int gdrom_check_drive(void)
{
  /*
   * check the state and disc type of the gdrom
   * 
   * note: I knew where to gather the information to get this done,
   *       but I really don't know how (or why) it's working in detail ;-)
   * 
   * returns: type of disc (0 AUDIO,0x1000 ROM,0x2000 XA,0x3000 CDI,0x8000 GD)
   *          or drive status (0 busy, 1 paused, 2 standby, 3 playing,
   *                           4 seeking, 5 scanning, 6 open, 7 no disc)
   */
  
  /* borrowed and modified from libdream's cdrom.c */
  uint32 *syscall_bc = (uint32*)0x8c0000bc;
  int (*syscall)() = (int (*)())(*syscall_bc); 
  int status[4], dummy[4];

  /* doing a call r6=0, r7=4 (why they are argv[2] and argv[3] I don't know
   * to get the status of the lid (6=open) according to Marcus info */
  syscall(&status, &dummy, 0, 4);

  return(status[0] | status[1] << 8);
}


void gdrom_check_lid(void)
{
  /*
   * check if the lid of the DC is open; if it's open, jump to menu
   * 
   * note: I knew where to gather the information to get this done,
   *       but I really don't know how (or why) it's working in detail ;-)
   */
  int c;
  
  if((gdrom_check_drive() & 0x0f) != 6)
    return;
  
  vid_clear(0,0,0);
  vid_init(vid_check_cable(),DM_320x240,PM_RGB555);
  DC_screen_size_X=320;
  DC_screen_size_Y=240;
  printxy("See ya!",132,116,0xf800,0);
  snd_init();
  for(c=0;c<90;c++) vid_waitvbl();
  /* borrowed from kos' kosh */
  (*(void(**)())0x8c0000e0)(1);
  /* thanks for the magic, Dan */
}


int load_file(const char *filename,void **buffer)
{
  /*
   * malloc the mem needed to load the file into ram and load it in there
   * 
   * filename: name of file to load
   * 
   * returns pointer to where file has been loaded
   */
  uint32 fd;
  int bytes_to_read,bytes_read;
  long size;
  void *bufferp;

#if (DEBUG == TRUE)
  printf("searching for %s\n",filename);
#endif
  /* no name, no load */
  if(strlen(filename) == 0)
  {
    return 0;
  }
  
  /* open the file */
  fd=iso_open(filename,O_RDONLY);
//  printf("file name: `%s\' -> %d\n",filename,fd);
  if(fd == 0)
  {
    /* file not found: return size of 0 */
    return fd;
  }
  
  /* get size of file */
  size=iso_seek(fd,0,SEEK_END);
#if (DEBUG == TRUE)
  printf("loading %08x (%d)\n",size,size);
#endif
  
  /* jump back to start of file */
  iso_seek(fd,0,SEEK_SET);
  
  /* allocate mem? */
  if(*buffer == NULL)
  {
    *buffer = (void*)malloc((size_t)size);
//    printf("malloc(%08x) -> %08x\n",size,buffer);
  }
  bufferp=*buffer;
  
  /* read the file to memory */
  for(bytes_to_read = size; bytes_to_read > 0; bytes_to_read-= bytes_read)
  {
    bytes_read = iso_read(fd,bufferp,bytes_to_read);
    bufferp+= bytes_read;
//    printf("bytes read: %08x (%d)\n",bytes_read,bytes_read);
  }
  iso_close(fd);

  return size;
}


void load_config(const char *filename)
{
  /*
   * load the configuration file (dcstella.ini) from cd
   * and create simple hash table
   * 
   * filename: name of ini-file
   */
  
  int fd,phase,size;
  char *buffer;
  char *key,*value,*c;

  size=load_file(filename,(void**)&buffer);

  key=buffer;
  value=buffer;
  numkeys=0;
  
  for(phase=1,c=buffer;(c-buffer) < size;c++)
  {
    switch(phase)
    {
     case 1:
      // skip spaces and tabs
      if((*c == ' ') || (*c == '\t')) key=c+1;
      else phase=2;
      break;
     case 2:
      // convert keyname to lower case
      *c=tolower(*c);
      // look for end of key
      if((*c == ' ') || (*c == '\t'))
      {
	*c = '\0';
	phase=3;
      }
      // fall through
     case 3:
      // look for '=' as delimiter beween key and value
      if(*c == '=')
      {
	*c = '\0';
	value=c+1;
	phase=4;
      }
      break;
     case 4:
      // skip spaces and tabs at the start of value
      if((*c == ' ') || (*c == '\t')) value=c+1;
      else phase=5;
      break;
     case 5:
      // look for end of key
      if(*c == '\n')
      {
	*c = '\0';
	phase=1;
	config_set_key_value(key,value);
	key=c+1;
      }
     default:
      break;
    }
    // search for beginning of line;
    if(*c == '\n')
    {
      key=c+1;
      phase=1;
    }
  }
}



int config_pos_by_key(const char *key)
{
  /*
   * get the position of a hash key in array
   * 
   * key: hash key
   * returns position or negative if not found
   */
  
  int i;
  for(i=0;i<numkeys;i++)
  {
    if(!strcmp(configkey[i],key))
    {
      return i;
    }
  }
  return -(i+1);
}



char* config_value_by_key(const char *key)
{
  /*
   * get value by hash key
   * 
   * key: hash key
   * returns hash value or NULL if not found
   */
  
  int i=config_pos_by_key(key);
//  printf("accessing key #%d (%s)\n",i,key);
  if(i >= 0)
    return configvalue[i];
  else
    return (char*)&nullstr;
}


void config_set_key_value(const char *key, const char *value)
{
  /*
   * create new or replace data by hash key
   * 
   * key: hash key
   * value: hash value
   */
  
  int pos;
  
  pos=config_pos_by_key(key);
  if(pos >= 0)
  {
    free(configvalue[pos]);
    configvalue[pos]=(char*)strdup(value);
  }
  else
  {
    pos=-pos-1;
    if(pos <= MAX_KEYS)
    {
      configkey[numkeys]  =(char*)strdup(key);
      configvalue[numkeys]=(char*)strdup(value);
      numkeys++;
    }
  }
}


void generate_roms_list(const char *dirname,const char *listname)
{
  /*
   * read dir into arrays
   * 
   * dirname: name of directory containing roms
   * listname: name of listfile (rom_list.txt)
   */
  
  char buffer[256];
  char *filebuffer=NULL;
  dirent_t *dirent;
  int  size;
  int  i=0,n=0,m=0,skip=0;
  char *c;
  
  if(listname)
  {
    size=load_file(listname,(void**)&filebuffer);
  }
  else
  {
    size=0;
  }

  if(size) // use file
  {
    char *file, *name;
    
    file=filebuffer;
    name=NULL;
    
    for(i=0,c=filebuffer;c<(filebuffer+size);c++)
    {
      if(!name)
      {
	if(*c == ' ')
	{
	  *c = '\0';
	  name = c+1;
	}
      }
      
      if(*c == '\n')
      {
	if(name)
	{
	  *c = '\0';
	  sprintf(buffer,"%s/%s",dirname,file);
	  filename[i]=(char*)strdup(buffer);
	  sprintf(buffer,"%s",name);
	  romname[i]=(char*)strdup(buffer);
	}
	
	file=c+1;
	name=NULL;
	i++;
      }
      if(i >= MAX_ROMS-1) break;
    }
    free(filebuffer);
  }
  
  uint32 fd;
  fd=iso_open(dirname, (O_RDONLY | O_DIR));
  if(fd)
  {
    for(n=i;i<MAX_ROMS;n++)
    {
      dirent=iso_readdir(fd);
      if(!dirent) break;
      sprintf(buffer,"%s/%s",dirname,dirent->name);
      skip = 0;
      for(m = 0; m < i; m++)
      {
	if(!strcasecmp(buffer,filename[m])) skip = 1;
      }
      if(skip) continue;
      filename[i]=(char*)strdup(buffer);
      sprintf(buffer,"%s",dirent->name);
      for(c=buffer;*c;c++)
      {
	if(*c == '_') *c = ' ';
      }
      romname[i]=(char*)strdup(buffer);
      i++;
    }
  }
  if(fd)
  {
    iso_close(fd);
  }
  
  romname[i] = NULL;
  filename[i] = NULL;
}


char *get_romname(int number)
{
  /*
   * get the name of the rom (just a wrapper function)
   *
   * number: number in array
   * returns name of rom
   */
  
  return romname[number];
}

char *get_filename(int number)
{
  /*
   * get the filename of the rom (just a wrapper function)
   *
   * number: number in array
   * returns filename of rom
   */

  return filename[number];
}
